import serial
import serial.tools.list_ports
import os, time, logging
from typing_extensions import Self
from .hardware_abc import HardwareABC, HardwarePort, OpeningPortException, TimeoutException

logger = logging.getLogger(__name__)

class KLineHardware(HardwareABC):
	'''
	Hardware class for serial devices, using pyserial as a backend
	'''

	def __init__ (self, port: str, baudrate: int = 10400, timeout: float = 2) -> None:
		self.port, self.baudrate = port, baudrate
		self.timeout = timeout

	def open (self) -> bool:
		try:
			self.socket = serial.Serial(self.port, self.baudrate, timeout=self.timeout)
		except serial.serialutil.SerialException as e:
			raise OpeningPortException(e)
		
		self._reset_adapter()
		self._set_kline_mode()

		return True
	
	def read (self, length: int) -> bytes:
		message = self.socket.read(length)

		if (len(message) < length):
			raise TimeoutException
		return message

	def write (self, data: bytes) -> int:
		# @todo - this timeout is needed for now, otherwise 
		# we'll get a timeout when trying to start diagnostic session
		# find out _why_ is it needed - some magic value instead 
		# of something calculated shouldnt be required here
		time.sleep((50/1000)*2)
		bytes_written = self.socket.write(data)

		while self.socket.out_waiting > 0:
			time.sleep(0.001)

		echo = self.socket.read(bytes_written)
		if (echo != data):
			logger.error('K-Line echo different than sent payload! \nPayload: {}\nEcho: {}'.format(
				' '.join([hex(x) for x in list(data)]),
				' '.join([hex(x) for x in list(echo)])
			))

		return bytes_written

	def close (self) -> None:
		try:
			self.socket.break_condition = False
			self.socket.close()
		except AttributeError:
			pass

	def set_timeout (self, timeout: float) -> Self:
		self.socket.timeout = timeout
		self.timeout = timeout
		return self

	def set_baudrate (self, baudrate: int) -> Self:
		self.socket.baudrate = baudrate
		self.baudrate = baudrate
		self.socket.flush()
		return self

	def _reset_adapter (self) -> None:
		'''
		Flip the Data Terminal Ready state. 
		This is required on some of the knock-off adapters which will otherwise fail to perform Fast Init
		'''
		self.socket.setDTR(1)
		time.sleep(0.1)
		self.socket.setDTR(0)
		time.sleep(0.1)

	def _set_kline_mode (self) -> None:
		'''
		Set the adapter to KLine (KKL) mode by flipping DTR and RTS to 0.
		This is required on some of the knock-off adapters which will otherwise fail to perform Fast Init
		'''
		self.socket.setDTR(0)
		self.socket.setRTS(0)
		time.sleep(0.1)

	def iso14230_fast_init (self, payload: bytes) -> tuple[bytes, int, int]:
		'''
		Perform FastInit by bringing the bus down for 25ms and then up for 25ms followed by a payload

		:return: response (1 byte), elapsed time in low position, elapsed time in high position
		'''

		# FastInit low
		start_time = time.perf_counter()
		self.socket.break_condition = True
		self.socket.flush()  # Ensure the break is sent immediately
		while (time.perf_counter() - start_time) < 0.025:
			pass  # Busy-wait until 25 ms has passed
		elapsed_time_low = time.perf_counter() - start_time

		# FastInit high
		start_time = time.perf_counter()
		self.socket.break_condition = False
		self.socket.flush()  # Ensure the break is sent immediately
		while (time.perf_counter() - start_time) < 0.025:
			pass  # Busy-wait until 25 ms has passed
		elapsed_time_high = time.perf_counter() - start_time
		
		# Send payload and read response
		self.socket.write(bytes(payload))
		response = self.socket.read(1) # most ECUs will respond with 0x00 after a successful init
		self.socket.read(len(payload)) # i have no idea why this is suddenly returning an echo
		return response, elapsed_time_low, elapsed_time_high

	@staticmethod
	def available_ports () -> list[HardwarePort]:
		devices = []
		for port in reversed(serial.tools.list_ports.comports()):
			devices.append(HardwarePort(port.device, port.name, port.manufacturer, port.serial_number))
		return devices


'''
Transport layers
'''

from .kwp2000_over_kline_transport import Kwp2000OverKLineTransport
from .kwp2000_over_can_transport import Kwp2000OverCanTransport
from .transport_abc import TransportABC, RawPacket, PacketDirection

__all__ = ['TransportABC', 'RawPacket', 'PacketDirection', 'Kwp2000OverKLineTransport', 'Kwp2000OverCanTransport']



from typing_extensions import Self
from typing import Union

class Kwp2000Command:
	service_identifier: int = 0x0

	def __init__ (self, data: Union[bytes, list[int]]=bytes()) -> None:
		if (isinstance(data, list)): # @todo: deprecate?
			data = bytes(data)

		self.data: bytes = data

	def set_data (self, data: Union[bytes, list[int]]=bytes()) -> Self:
		if (isinstance(data, list)): # @todo: deprecate?
			data = bytes(data)

		self.data = data
		return self

	def get_data (self, as_list=False) -> bytes:
		if as_list:
			return list(self.data)
		return self.data

	def append_data (self, data: Union[bytes, list[int]]=bytes()) -> Self:
		if (isinstance(data, list)): # @todo: deprecate?
			data = bytes(data)

		self.data += data
		return self

	def set_service_identifier (self, service_identifier: int) -> Self:
		self.service_identifier = service_identifier
		return self

	def get_service_identifier (self) -> int:
		return self.service_identifier

	def __str__ (self) -> str:
		return 'KWPCommand(sid={}, data={})'.format(hex(self.get_service_identifier()), ' '.join([hex(x) for x in self.get_data(as_list=True)]))

	def __repr__ (self) -> str:
		return self.__str__()

class Kwp2000CommandWithSubservices(Kwp2000Command):
	def set_subservice_identifier (self, subservice_identifier: int) -> Self:
		'''
		Overwrite the first byte of data with new subservice identifier
		'''
		return self.set_data(bytes([subservice_identifier]) + self.get_data()[1:])

	def get_subservice_identifier (self) -> int:
		return self.get_data()[0]

	def get_data (self, as_list: bool = False) -> bytes:
		'''
		Get command payload, with subservice identifier being the first byte
		'''
		if as_list: # @todo: deprecate?
			return list(self.data)

		return self.data

	def __str__ (self) -> str:
		return 'KWPSubservice(sid={}, sub={}, data={})'.format(
			hex(self.get_service_identifier()), 
			hex(self.get_subservice_identifier()),
			' '.join([hex(x) for x in self.get_data(as_list=True)])
		)



'''
Hardware layers - concrete implementations of devices
'''

from .kline_hardware import KLineHardware
from .can_hardware import CanHardware
from .hardware_abc import HardwareABC, HardwarePort, HardwareException, OpeningPortException, SendingException, ReadingException, TimeoutException

__all__ = ['HardwareABC', 'KLineHardware', 'CanHardware', 'HardwarePort', 'HardwareException', 'OpeningPortException', 'SendingException', 'ReadingException', 'TimeoutException']



import struct, time
from sys import platform
from scapy.config import conf
if platform.startswith('win32'):
	conf.contribs['ISOTP'] = {'use-can-isotp-kernel-module': False}
else:
	conf.contribs['ISOTP'] = {'use-can-isotp-kernel-module': True}
from scapy.contrib.isotp import *
from .transport_abc import TransportABC, RawPacket, PacketDirection
from ..hardware.hardware_abc import HardwareABC, TimeoutException


class Kwp2000OverCanTransport (TransportABC):
	def __init__ (self, hardware: HardwareABC, tx_id: hex, rx_id: hex) -> None:
		self.hardware = hardware
		self.tx_id, self.rx_id = tx_id, rx_id
		self.isotp = ISOTPSocket(self.hardware.socket, tx_id, rx_id, padding=True)
		
	def send_pdu (self, pdu: bytes) -> int:
		data = pdu
		bytes_written = self.isotp.send(data)

		self.buffer_push(RawPacket(direction=PacketDirection.OUTGOING, data=data, timestamp=int(time.time()*1000)))

		return len(data) # isotp socket doesnt return how many bytes were written

	def read_pdu (self) -> bytes:
		try:
			frame = self.isotp.sniff(timeout=self.hardware.get_timeout(), count=1)[0]
		except IndexError:
			self.isotp.close() # close the background thread that sends flow control frames
			raise TimeoutException

		self.buffer_push(RawPacket(direction=PacketDirection.INCOMING, data=frame.data, timestamp=int(time.time() * 1000)))

		return frame.data

	def send_read_pdu (self, data: bytes) -> bytes:
		self.buffer_push(RawPacket(direction=PacketDirection.OUTGOING, data=data, timestamp=int(time.time()*1000)))
		response = self.isotp.sr1(ISOTP(bytes(data)), verbose=False)
		self.buffer_push(RawPacket(direction=PacketDirection.INCOMING, data=response.data, timestamp=int(time.time() * 1000)))

		return response.data


from abc import ABC
from ..transport.transport_abc import TransportABC

class ProtocolABC (ABC):
	def __init__ (self, transport: TransportABC) -> None:
		self.transport = transport

	def open (self) -> bool:
		'''
		Initialize the underlaying layers - get ready to interact
		'''
		pass

	def close (self) -> None:
		'''
		Shutdown the underlaying layers
		'''
		pass

	def __del__ (self) -> None:
		self.close()


import os, math
from sys import platform
from typing_extensions import Self
from scapy.config import conf
if platform.startswith('win32'):
	conf.contribs['CANSocket'] = {'use-python-can': True}
else:
	conf.contribs['CANSocket'] = {'use-python-can': False}
	from scapy.contrib.cansocket import CANSocket
from scapy.layers.can import CAN, CAN_MTU, CAN_MAX_DLEN
from scapy.error import Scapy_Exception
from .hardware_abc import HardwareABC, HardwarePort, OpeningPortException, SendingException, TimeoutException

CAN_HEADER_LEN = CAN_MTU-CAN_MAX_DLEN

class CanHardware(HardwareABC):
	'''
	Hardware class for CAN Bus interfaces. Uses Scapy as a backend
	'''

	def __init__ (self, port: str, tx_id: int = None, rx_id: int = None, timeout: int = 1) -> None:
		self.port = port
		self.tx_id, self.rx_id = tx_id, rx_id
		self.timeout = timeout

	def _build_filters (self) -> list[dict]:
		filters = []
		
		if self.rx_id:
			filters.append({'can_id': self.rx_id, 'can_mask': 0x7FF})
		else:
			filters = None

		return filters

	def open (self) -> bool:
		try:
			self.socket = CANSocket(channel=self.port, can_filters=self._build_filters())
		except (OSError, Scapy_Exception) as e:
			raise OpeningPortException(e)
		return True

	def read (self, length: int) -> bytes:
		try:
			packet = self.socket.sniff(timeout=self.timeout, count=1)[0]
		except IndexError:
			raise TimeoutException

		return packet.data

	def write (self, data: bytes, tx_id: int = None) -> int:
		tx_id = tx_id if tx_id != None else self.tx_id
		if tx_id == None:
			raise SendingException('Tried to send without CAN Bus ID specified')

		packet = CAN(identifier=tx_id, data=data)

		bytes_written = self.socket.send(packet)

		return bytes_written-CAN_HEADER_LEN

	def close (self) -> None:
		self.socket.close()

	def get_baudrate (self) -> int:
		raise NotImplementedError

	def set_baudrate (self, baudrate: int) -> Self:
		raise NotImplementedError

	def set_timeout (self, timeout: float) -> Self:
		self.socket.timeout = timeout
		self.timeout = timeout
		return self

	@staticmethod
	def _available_ports_linux () -> list[str]:
		'''
		Returns a list of available CAN network interfaces by checking the /sys/class/net/<iface>/type file

		:return: a list containing interface identifiers (can0, vcan0, ...)
		'''
		NET_DIR = '/sys/class/net'
		interfaces = []

		if not os.path.exists(NET_DIR):
			return interfaces

		for iface in os.listdir(NET_DIR):
			type_path = os.path.join(NET_DIR, iface, 'type')
			if os.path.isfile(type_path):
				with open(type_path, 'r') as f:
					iface_type = f.read().strip()

				if iface_type == '280':
					interfaces.append(iface)

		return interfaces

	@staticmethod
	def available_ports () -> list[HardwarePort]:
		if platform.startswith('win32'):
			return []
		else:
			ports = [HardwarePort(port=x, port_name=x) for x in CanHardware._available_ports_linux()]
		
		return ports
